from ._b2d import *
from .tools import _classExtender
from . extend_math import vec2

class ParticleGroupFlag(object):
    # prevents overlapping or leaking.
    solidParticleGroup = 1 << 0
    # Keeps its shape.
    rigidParticleGroup = 1 << 1
    # Won't be destroyed if it gets empty.
    particleGroupCanBeEmpty = 1 << 2
    # Will be destroyed on next simulation step.
    particleGroupWillBeDestroyed = 1 << 3
    # Updates depth data on next simulation step.
    particleGroupNeedsUpdateDepth = 1 << 4
    particleGroupInternalMask = particleGroupWillBeDestroyed | particleGroupNeedsUpdateDepth

class ParticleFlag(object):
    waterParticle = 0
    # Removed after next simulation step.
    zombieParticle = 1 << 1
    # Zero velocity.
    wallParticle = 1 << 2
    # With restitution from stretching.
    springParticle = 1 << 3
    # With restitution from deformation.
    elasticParticle = 1 << 4
    # With viscosity.
    viscousParticle = 1 << 5
    # Without isotropic pressure.
    powderParticle = 1 << 6
    # With surface tension.
    tensileParticle = 1 << 7
    # Mix color between contacting particles.
    colorMixingParticle = 1 << 8
    # Call b2DestructionListener on destruction.
    destructionListenerParticle = 1 << 9
    # Prevents other particles from leaking.
    barrierParticle = 1 << 10
    # Less compressibility.
    staticPressureParticle = 1 << 11
    # Makes pairs or triads with other particles.
    reactiveParticle = 1 << 12
    # With high repulsive force.
    repulsiveParticle = 1 << 13
    # Call b2ContactListener when this particle is about to interact with
    # a rigid body or stops interacting with a rigid body.
    # This results in an expensive operation compared to using
    # fixtureContactFilterParticle to detect collisions between
    # particles.
    fixtureContactListenerParticle = 1 << 14
    # Call b2ContactListener when this particle is about to interact with
    # another particle or stops interacting with another particle.
    # This results in an expensive operation compared to using
    # particleContactFilterParticle to detect collisions between
    # particles.
    particleContactListenerParticle = 1 << 15
    # Call b2ContactFilter when this particle interacts with rigid bodies.
    fixtureContactFilterParticle = 1 << 16
    # Call b2ContactFilter when this particle interacts with other
    # particles.
    particleContactFilterParticle = 1 << 17

def particle_system_def(
    strict_contact_check = None,
    density = None,
    gravity_scale = None,
    radius = None,
    max_count = None,
    pressure_strength = None,
    damping_strength = None,
    elastic_strength = None,
    spring_strength = None,
    viscous_strength = None,
    surface_tension_pressure_strength = None,
    surface_tension_normal_strength = None,
    repulsive_strength = None,
    powder_strength = None,
    ejection_strength = None,
    static_pressure_strength = None,
    static_pressure_relaxation = None,
    static_pressure_iterations = None,
    color_mixing_strength = None,
    destroy_by_age = None,
    lifetime_granularity = None,
):
    d = ParticleSystemDef()

    if strict_contact_check  is not None:
        d.strict_contact_check = strict_contact_check
    if density  is not None:
        d.density = density
    if gravity_scale  is not None:
        d.gravity_scale = gravity_scale
    if radius  is not None:
        d.radius = radius
    if max_count  is not None:
        d.max_count = max_count
    if pressure_strength  is not None:
        d.pressure_strength = pressure_strength
    if damping_strength  is not None:
        d.damping_strength = damping_strength
    if elastic_strength  is not None:
        d.elastic_strength = elastic_strength
    if spring_strength  is not None:
        d.spring_strength = spring_strength
    if viscous_strength  is not None:
        d.viscous_strength = viscous_strength
    if surface_tension_pressure_strength  is not None:
        d.surface_tension_pressure_strength = surface_tension_pressure_strength
    if surface_tension_normal_strength  is not None:
        d.surface_tension_normal_strength = surface_tension_normal_strength
    if repulsive_strength  is not None:
        d.repulsive_strength = repulsive_strength
    if powder_strength  is not None:
        d.powder_strength = powder_strength
    if ejection_strength  is not None:
        d.ejection_strength = ejection_strength
    if static_pressure_strength  is not None:
        d.static_pressure_strength = static_pressure_strength
    if static_pressure_relaxation  is not None:
        d.static_pressure_relaxation = static_pressure_relaxation
    if static_pressure_iterations  is not None:
        d.static_pressure_iterations = static_pressure_iterations
    if color_mixing_strength  is not None:
        d.color_mixing_strength = color_mixing_strength
    if destroy_by_age  is not None:
        d.destroy_by_age = destroy_by_age
    if lifetime_granularity  is not None:
        d.lifetime_granularity = lifetime_granularity

    return d


def particle_group_def(
    flags=None,
    group_flags=None,
    position=None,
    angle=None,
    linear_velocity=None,
    angular_velocity=None,
    color=None,
    strength=None,
    shape=None,
    stride=None,
    particle_count=None,
    lifetime=None,
    group=None
):
    d = ParticleGroupDef()
    if flags is not None:
        d.flags = flags
    if group_flags is not None:
        d.group_flags = group_flags
    if position is not None:
        d.position = position
    if angle is not None:
        d.angle = angle
    if linear_velocity is not None:
        d.linear_velocity = linear_velocity
    if angular_velocity is not None:
        d.angular_velocity = angular_velocity
    if color is not None:
        d.color = color
    if color is not None:
        d.color = color
    if strength is not None:
        d.strength = strength
    if shape is not None:
        d.shape = shape
    if stride is not None:
        d.stride = stride
    if particle_count is not None:
        d.particle_count = particle_count
    if lifetime is not None:
        d.lifetime = lifetime
    if group is not None:
        d.group(shape)
    
    return d


def particle_def(
    flags = None,
    position = None,
    velocity = None,
    color = None,
    lifetime = None,
    group = None
):
    pd = ParticleDef() 

    if flags is not None:
        pd.flags = flags
    if position is not None:
        pd.position = vec2(position)
    if velocity is not None:
        pd.velocity = vec2(velocity)
    if color is not None:
        pd.color = color
    if lifetime is not None:
        pd.lifetime = lifetime
    if group is not None:
        pd.group = group

    return pd
#class _ParticleSystem(b2ParticleSystem):
#    pass
#_classExtender(_ParticleSystem,['group','shape'])


class _ParticleGroupDef(ParticleGroupDef):

    @property
    def group(self):
        return self._group()
    @group.setter
    def group(self, group):
        self._setGroup(shape)


    @property
    def shape(self):
        return self._shape()
    @shape.setter
    def shape(self, shape):
        self._setShape(shape)

_classExtender(_ParticleGroupDef,['group','shape'])





import numpy as np
from time import sleep
from threading import Event, Thread, Lock
from ipycanvas import Canvas, hold_canvas
from ipywidgets import Label, HTML, Button, HBox, VBox
import functools
import IPython
import time
from ipycanvas import Canvas,MultiCanvas, hold_canvas
import ipywidgets
from . jupyter_batch_debug_draw import JupyterBatchDebugDraw


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a-b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)

def scale_color(color):
    return [float(c)/255.0 for c in color]

def rgb(color):
    r = int(color[0])
    g = int(color[1])
    b = int(color[2])
    return f"rgb({r},{g},{b})"


_id_to_gui = dict()


class JupyterGui(object):
    def __init__(self, testbed_cls, settings, testbed_kwargs=None):
        
        self.id = settings["id"]
        if self.id in _id_to_gui:
            old_self = _id_to_gui[self.id]
            old_self._terminate()
        _id_to_gui[self.id] = self

        # settings
        resolution = settings.get("resolution",  (1024,768))
        self.resolution = resolution

        self.settings = settings
        # steping settings
        self._fps = settings.get("fps", 30) 
        self._dt_s = 1.0 / self._fps

        # testworld
        if testbed_kwargs is None:
            testbed_kwargs = dict()
        self.testbed_kwargs = testbed_kwargs
        self.testbed_cls = testbed_cls
        self._testworld  = None

        # debug_draw
        self.debug_draw = None
        self.flip_bit = False
        self._debug_draw_flags = settings.get("draw_flags", ['shape','joint','particle'])

        # flag to stop loop
        self._exit = False
       
        self.scale = settings.get("scale", 20) 
        self.translate = settings.get("translate", (10,10))


        # Thread related
        # events
        self.paused = Event()
        self.reached_end = Event()
        self._world_lock = Lock()

    def _terminate(self):
        if not self.paused.isSet():
            self.paused.set()


    def make_testworld(self):

        if self._testworld is not None:
            self._testworld.say_goodbye_world()
        self._testworld = self.testbed_cls(**self.testbed_kwargs)

        # make debug draw
        self.debug_draw = JupyterBatchDebugDraw(self.multi_canvas[self.flip_bit], 
            flags=self._debug_draw_flags)
        self.debug_draw.screen_size = self.resolution
        self.debug_draw.scale = self.scale
        self.debug_draw.translate = self.translate
        self.debug_draw.flip_y = True
        self._testworld.set_debug_draw(self.debug_draw)

    def start_ui(self):
        # make the canvas
        self.multi_canvas = MultiCanvas(n_canvases = 2, 
        width=self.resolution[0], height=self.resolution[1])
        self.out = ipywidgets.Output()
        self.flip_bit = False


        # _setup_ipywidgets_gui
        self._setup_ipywidgets_gui()



        #make the world
        self.make_testworld()

        def on_mouse_down( xpos, ypos):
            # with self.out:
            # self.multi_canvas[1].fill_circle(xpos, ypos, 10)
            # self.multi_canvas[1].flush()
            pos = self.debug_draw.screen_to_world((xpos, ypos))
            pos = pos.x, pos.y
            with self._world_lock:
                self._testworld.on_mouse_down(pos)
 


        # moue callbacks
        def on_mouse_up( xpos, ypos):
            pos = self.debug_draw.screen_to_world((xpos, ypos))
            pos = pos.x, pos.y
            with self._world_lock:
                self._testworld.on_mouse_up(pos)


        def on_mouse_move( xpos, ypos):
            # self.multi_canvas[1].fill_circle(xpos, ypos, 10)
            # self.multi_canvas[1].flush()
            # self.multi_canvas[0].fill_circle(xpos, ypos, 10)
            # self.multi_canvas[0].flush()
            pos = self.debug_draw.screen_to_world((xpos, ypos))
            pos = pos.x, pos.y
            with self._world_lock:
                self._testworld.on_mouse_move(pos)

        self.multi_canvas[1].on_mouse_down(on_mouse_down)
        self.multi_canvas[1].on_mouse_up(on_mouse_up)
        self.multi_canvas[1].on_mouse_move(on_mouse_move)



    
        target_fps = 30
        dt_desired_ms = (1.0/target_fps)*1000.0
        dt_desired_s = dt_desired_ms/1000.0
        



        for ci in range(2):
            self.multi_canvas[ci].line_width = self.settings.get("line_width", 1)
            # self.multi_canvas[ci].scale(self.scale, -1.0*self.scale)
            # self.multi_canvas[ci].line_width = 1.0 / self.scale



        Thread(target=self._loop).start() # Start it by default


    def _loop(self):
        if self.reached_end.isSet():
            self.reached_end.clear()
        # Event loop
        while not self.paused.isSet():
            
            t0 = time.time()
            if self._exit:
                break

            self._single_step()
            t1 = time.time()

            delta = t1 - t0
            if delta < self._dt_s:
                time.sleep(self._dt_s - delta)
        self.reached_end.set()


    def _setup_ipywidgets_gui(self):
        # buttons
        start_btn =         Button(icon='play')
        step_forward_btn =  Button(icon='step-forward')
        step_forward_btn.disabled = True
        pause_btn =         Button(icon='pause')
        reset_btn =         Button(icon='stop')

        # sliders speed / fps
        fps_slider = ipywidgets.IntSlider(
            value=self._fps,
            min=1,
            max=100,
            step=1
        )

        speed_slider = ipywidgets.FloatSlider(
            value=1.0,
            min=0.1,
            max=10.0,
            step=0.1
        )


        def pause(btn=None):
            if not self.paused.isSet():
                step_forward_btn.disabled = False
                self.paused.set()
        pause_btn.on_click(pause)

        def start(btn=None):
            step_forward_btn.disabled = True
            if self.paused.isSet():
                self.paused.clear()
            if self.reached_end.isSet():
                self.reached_end.clear()
                Thread(target=self._loop).start() 
        start_btn.on_click(start)

        def step_forward(btn=None):
            self._single_step()
        step_forward_btn.on_click(step_forward)

        def reset(btn):
            pause()
            while not self.reached_end.wait(0.02):
                pass
            self.make_testworld()
            self._single_step()
            # start()
        reset_btn.on_click(reset)


        draw_checkboxes = dict(
            shapes=ipywidgets.Checkbox(value=True),
            joints=ipywidgets.Checkbox(value=True),
            aabb=ipywidgets.Checkbox(value=False),
            com=ipywidgets.Checkbox(value=False),
            pairs=ipywidgets.Checkbox(value=False)
        )   

        def on_flag_change(v, flag):
            v = v['new']
            if v :
                self.debug_draw.append_flags(flag)
            else:
                self.debug_draw.clear_flags([flag])

            if self.paused.isSet():
                self._draw_world(self.debug_draw._canvas)

        items = []
        flags = ['shape','joint','aabb','pair','center_of_mass','particle']
        for f in flags:
            label = ipywidgets.Label(value=f'Draw {f} :')
            checkbox = ipywidgets.Checkbox(value=bool(f in self._debug_draw_flags))
            checkbox.observe(functools.partial(on_flag_change, flag=f), names='value')
            items.append(label)
            items.append(checkbox)


        draw_flags = ipywidgets.GridBox(items, layout=ipywidgets.Layout(grid_template_columns="repeat(4, 200px)"))


        # display
        IPython.display.display(self.out)
        with self.out:
            IPython.display.display(self.multi_canvas, 
                VBox([
                    HBox([start_btn,  step_forward_btn,  pause_btn, reset_btn]),
                    draw_flags
                ])
            )

    def _single_step(self):
        self._step_world()

        canvas = self.multi_canvas[self.flip_bit]
        self.flip_bit = not self.flip_bit
        next_canvas = self.multi_canvas[self.flip_bit]
        assert canvas != next_canvas
        with hold_canvas(next_canvas):
            self.debug_draw._canvas = next_canvas
            self._draw_world(next_canvas)

        # clear this one
        canvas.clear()
                

    def _step_world(self):
        with self._world_lock:
            self._testworld.step(self._dt_s)

    def _draw_world(self, canvas):
        old_style = canvas.fill_style
        canvas.fill_style = 'black'
        canvas.fill_rect(0,0, self.resolution[0],self.resolution[1])

        self._testworld.world.draw_debug_data()
        canvas.fill_style = old_style


from . _b2d import DrawCaller
from . tools import _classExtender, GenericB2dIter
from contextlib import contextmanager


import numbers 

class DrawFlags(object):
    shape_bit              = 0x0001
    joint_bit              = 0x0002
    aabb_bit               = 0x0004
    pair_bit               = 0x0008
    center_of_mass_bit     = 0x0010
    particle_bit           = 0x0020

draw_flags_dict = {
    "shape"              : 0x0001,
    "joint"              : 0x0002,
    "aabb"               : 0x0004,
    "pair"               : 0x0008,
    "center_of_mass"     : 0x0010,
    "particle"           : 0x0020
}



def _extendDrawCaller():
    
    def append_flags(self, flag_list_or_int):
        if isinstance(flag_list_or_int, numbers.Number):
            self._append_flags_int(flag_list_or_int)
        else:
            flag_list = flag_list_or_int
            if isinstance(flag_list, str):
                flag_list = [flag_list]
            for flag in flag_list:
                self._append_flags_int(draw_flags_dict[flag])
    DrawCaller.append_flags =append_flags

    def clear_flags(self, flag_list_or_int):
        if isinstance(flag_list_or_int, numbers.Number):
            self._clear_flags_int(flag_list_or_int)
        else:
            flag_list = flag_list_or_int
            if isinstance(flag_list, str):
                flag_list = [flag_list]
            for flag in flag_list:
                self._clear_flags_int(draw_flags_dict[flag])

    DrawCaller.clear_flags = clear_flags

    
_extendDrawCaller()
del _extendDrawCaller




import numpy
import b2d as b2
import pygame
import pygame.locals 
import time

from .pygame_debug_draw import PyGameBatchDebugDraw

class PygameGui(object):
    def __init__(self, testbed_cls, settings=None, testbed_kwargs=None):

        # settings
        resolution = settings.get("resolution", (1024,768))
        self.settings = settings
        self.resolution = resolution
        print("resolution",resolution)

        # testworld
        if testbed_kwargs is None:
            testbed_kwargs = dict()
        self.testbed_kwargs = testbed_kwargs
        self.testbed_cls = testbed_cls
        self._testworld  = None

        # flag to stop loop
        self._exit = False

        # surface
        self._surface = None

        # mouse state
        self._last_was_drag = False
        self._last_pos = None

        # steping settings
        self._fps = settings.get("fps", 30) 
        self._dt_s = 1.0 / self._fps

     
    def make_testworld(self):

        if self._testworld is not None:
            self._testworld.say_goodbye_world()
        self._testworld = self.testbed_cls(**self.testbed_kwargs)




    def start_ui(self):

        # make the world
        self.make_testworld()

        # Initialise screen
        pygame.init()
        self._surface = pygame.display.set_mode(self.resolution)
        pygame.display.set_caption(self.testbed_cls.name)


        # debug draw
        # self.debug_draw = PygameDebugDraw(surface=self._surface)
        self.debug_draw = PyGameBatchDebugDraw(surface=self._surface)
        self.debug_draw.screen_size = self.resolution
        self.debug_draw.flip_y = True
        self.debug_draw.scale = self.settings.get("scale",20.0)
        self.debug_draw.translate = self.settings.get("translate", [10.0,10.0])
        self.debug_draw.append_flags([
            'shape',
            'joint',
            # 'aabb',
            # 'pair',
            'center_of_mass',
            'particle'
        ])
        self._testworld.set_debug_draw(self.debug_draw)

        # Event loop
        while 1:
            
            t0 = time.time()
            self._handle_events()

            if self._exit:
                break

            self._step_world()
            self._draw_world()
            t1 = time.time()

            delta = t1 - t0
            if delta < self._dt_s:
                time.sleep(self._dt_s - delta)

     

    def _zoom_in(self):
        self.debug_draw.scale *= 1.25

    def _zoom_out(self):
        self.debug_draw.scale *= 0.75

    def _handle_events(self):

        pressed_keys = pygame.key.get_pressed()
        pressed_mouse_buttons = pygame.mouse.get_pressed()


        ctrl_pressed = pressed_keys[pygame.K_LCTRL] or pressed_keys[pygame.K_RCTRL]

        drag_mode = ctrl_pressed and pressed_mouse_buttons[0]

        if(drag_mode and self._last_was_drag):
            pos = pygame.mouse.get_pos()
            delta = [
                self._last_pos[0] - pos[0],
                self._last_pos[1] - pos[1]
            ]
            translate = self.debug_draw.translate
            self.debug_draw.translate = (
                translate.x - delta[0],
                translate.y + delta[1]
            )
        # 
        self._last_was_drag = drag_mode
        self._last_pos = pygame.mouse.get_pos()

        # TRANSLATION
        if ctrl_pressed:
            d = 0.1
            if pressed_keys[pygame.K_UP]:
                t = self.debug_draw.translate
                self.debug_draw.translate = (t.x, t.y - d)
            elif pressed_keys[pygame.K_DOWN]:
                t = self.debug_draw.translate
                self.debug_draw.translate = (t.x, t.y + d)
            elif pressed_keys[pygame.K_LEFT]:
                t = self.debug_draw.translate
                self.debug_draw.translate = (t.x - d, t.y)
            elif pressed_keys[pygame.K_RIGHT]:
                t = self.debug_draw.translate
                self.debug_draw.translate = (t.x + d, t.y)

        for event in pygame.event.get():
            if event.type == pygame.locals.QUIT:
                self._exit = True
                break

            elif event.type == pygame.MOUSEBUTTONDOWN:
                if event.button == 1:
                    screen_pos = pos = pygame.mouse.get_pos()
                    world_pos = self.debug_draw.screen_to_world(screen_pos)
                    self._testworld.on_mouse_down(world_pos)


                # zoom
                if ctrl_pressed:

                    if event.button == 4:
                        self._zoom_in()
                    elif event.button == 5:
                        self._zoom_out()

            elif event.type == pygame.MOUSEBUTTONUP:
                if event.button == 1:
                    screen_pos = pos = pygame.mouse.get_pos()
                    world_pos = self.debug_draw.screen_to_world(screen_pos)
                    self._testworld.on_mouse_up(world_pos)

            elif event.type == pygame.MOUSEMOTION:
                screen_pos = pos = pygame.mouse.get_pos()
                world_pos = self.debug_draw.screen_to_world(screen_pos)
                self._testworld.on_mouse_move(world_pos)




    def _draw_world(self):
        self._surface.fill((0,0,0))
        self._testworld.draw_debug_data()
        pygame.display.update()

    def _step_world(self):
        self._testworld.step(self._dt_s)



import sys
import os


import os, sys


def is_notebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False    

def is_pyodide():
    return "pyodide" in sys.modules


def is_doc_build():
    return os.environ.get('READTHEDOCS') == 'True' or os.environ.get('PYB2D_SPHINX_BUILD') == 'True' 

def default_backend():
    if is_doc_build():
        from b2d.testbed.backend.matplotlib_gif_gui import MatplotlibGifGui
        return MatplotlibGifGui,{}
    elif is_pyodide():
        raise RuntimeError("pyodide has not yet a backend")
    elif is_notebook():
        from b2d.testbed.backend.jupyter import JupyterGui
        return JupyterGui,{}

    else:
        try:
            from b2d.testbed.backend.pygame import PygameGui
            return PygameGui,{}
        except:
            pass

        try:
            from b2d.testbed.backend.kivy import KivyGui
            return KivyGui,{}
        except:
            raise RuntimeError("no backend found: try installing pygame or kivy")



def run(example_cls, backend=None, gui_settings=None):

    if isinstance(backend, str):
        backend_name = backend
        backend_cls = None
    else:
        backend_name = None
        backend_cls = backend

    if backend_name is not None:
        if backend_name == "pygame":
            from b2d.testbed.backend.pygame import PygameGui
            backend_cls, default_gui_settings = PygameGui,{}
        elif backend_name == "kivy":
            from b2d.testbed.backend.kivy import KivyGui
            backend_cls, default_gui_settings = KivyGui,{}
        elif backend_name == "matplotlib_gif":
            from b2d.testbed.backend.matplotlib_gif_gui import MatplotlibGifGui
            backend_cls, default_gui_settings = MatplotlibGifGui,{}
        elif backend_name == "jupyter":
            from b2d.testbed.backend.jupyter import JupyterGui
            backend_cls, default_gui_settings = JupyterGui,{}

        if gui_settings is None:
            gui_settings = default_gui_settings


    if backend_cls is None:
        backend_cls, default_gui_settings = default_backend()
        if gui_settings is None:
            gui_settings = default_gui_settings

    if gui_settings is None:
        gui_settings = dict()

    return example_cls.run(backend_cls, gui_settings=gui_settings)


from abc import ABC
import numbers 


from . _b2d import *

from . tools import *
from . extend_math import *
from . extend_draw import *
from . extend_world import *
from . extend_body import *
from . extend_fixture import *
from . extend_shapes import *
from . extend_shapes import EdgeShape
from . extend_joints import *
from . extend_collision import *
from . extend_contact import *
from . query_callback import *
if BuildConfiguration.LIQUID_FUN:
    from . extend_particles import *
# from . destruction_listener import DestructionListener





class RayCastCallback(RayCastCallbackCaller):

    def __init__(self):
        super(RayCastCallback,self).__init__(self)

    def report_fixture(self, fixture, point, normal, fraction):
        raise NotImplementedError 
    #def report_particle(self, particleSystem, index):
    #    return False
    #def should_query_particle_system(self, particleSystem):
    #    return False


class ContactListener(ContactListenerCaller):

    def __init__(self):
        super(ContactListener,self).__init__(self)

    # def begin_contact(self, contact):
    #     pass

    # def end_contact(self, contact):
    #     pass

    # def begin_contact_particle_body(self, particleSystem, particleBodyContact):
    #     pass

    # def begin_contact_particle(self, particleSystem, indexA, indexB):
    #     pass

    # def end_contact_particle(self, particleSystem, indexA, indexB):
    #     pass

    # def pre_solve(self, contact, oldManifold):
    #     pass

    # def post_solve(self, contact, impulse):
    #     pass


        


class DebugDraw(DrawCaller):
    def __init__(self, float_colors=True):
        self.float_colors = float_colors
        super(DebugDraw, self).__init__(self, bool(float_colors))

    def begin_draw(self):
        pass

    def end_draw(self):
        pass

    def draw_solid_circle(self, center, radius, axis, c):
        raise NotImplementedError 

    def draw_point(self, center, size, c):
        raise NotImplementedError 

    def draw_circle(self, center, radius, c):
        raise NotImplementedError 

    def draw_segment(self,v1, v2, c):
        raise NotImplementedError 

    def draw_polygon(self,vertices, c):
        raise NotImplementedError 

    def draw_solid_polygon(self,vertices, c):
        raise NotImplementedError 

    def draw_particles(self, centers, radius,  c=None):
        raise NotImplementedError 
    
    def draw_transform(self, xf):
        raise NotImplementedError 








def batch_debug_draw_caller_cls(float_colors, float_coordinates, with_transform):

    color_type = ['uint8','float'][float_colors]
    coordinate_type = ['int32','float'][float_coordinates]

    cls_name = f"BatchDebugDrawCaller_{color_type}_{coordinate_type}_{with_transform}"

    return getattr(_b2d, cls_name)


def batch_debug_draw_cls(float_colors, float_coordinates, with_transform):

    base_cls = batch_debug_draw_caller_cls(float_colors, float_coordinates, with_transform)

    class BatchDebugDraw(base_cls):

        def __init__(self):
            super(BatchDebugDraw, self).__init__(self)

        def begin_draw(self):
            pass

        def end_draw(self):
            pass


        def draw_solid_polygons(self, points, connect, color):
            pass

        def draw_polygons(self, points, connect, color):
            pass

        def draw_segments(self, points, connect, color):
            pass

        def draw_solid_circles(self, centers, radii, axis, color):
            pass

        def draw_circles(self, centers, radii, color):
            pass

        def draw_particles(self, centers, radius, colors):
            pass


        def append_flags(self, flag_list_or_int):
            if isinstance(flag_list_or_int, numbers.Number):
                self._append_flags_int(flag_list_or_int)
            else:
                flag_list = flag_list_or_int
                if isinstance(flag_list, str):
                    flag_list = [flag_list]
                for flag in flag_list:
                    self._append_flags_int(draw_flags_dict[flag])

        def clear_flags(self, flag_list_or_int):
            if isinstance(flag_list_or_int, numbers.Number):
                self._clear_flags_int(flag_list_or_int)
            else:
                flag_list = flag_list_or_int
                if isinstance(flag_list, str):
                    flag_list = [flag_list]
                for flag in flag_list:
                    self._clear_flags_int(draw_flags_dict[flag])

    return BatchDebugDraw




class ContactFilter(ContactFilterCaller):

    def __init__(self):
        super(ContactFilter,self).__init__(self)


    def should_collide_fixture_fixture(self, fixtureA, fixtureB):
        pass

    def should_collide_fixture_particle(self, fixture, particleSystem, particleIndex):
        pass

    def should_collide_particle_particle(self, particleSystem, particleIndexA, particleIndexB):
        pass   



class DestructionListener(DestructionListenerCaller):

    def __init__(self):
        super(DestructionListener,self).__init__(self)

    def say_goodbye_joint(self, joint):
        pass
    def say_goodbye_fixture(self, fixture):
        pass
    def say_goodbye_particle_group(self, particleGroup):
        pass
    def say_goodbye_particle_system(self, particleSystem,index):
        pass


from types import MethodType

from . _b2d import *
from . tools import _classExtender, GenericB2dIter
from . extend_math import vec2
from . extend_shapes import *
from . extend_user_data import add_user_data_api


def body_def(btype=None,position=None,angle=None,linear_velocity=None,
            angular_velocity=None,linear_damping=None,angular_damping=None,
            allow_sleep=None, awake=None, fixed_rotation=None, bullet=None,
            user_data=None, int_user_data=None):
    d = BodyDef()
    if btype is not None:
        d.btype = btype
    if position is not None:
        d.position = vec2(position)
    if angle is not None:
        d.angle = angle
    if linear_velocity is not None:
        d.linear_velocity = vec2(linear_velocity)
    if angular_velocity is not None:
        d.angular_velocity = float(angular_velocity)
    if linear_damping is not None:
        d.linear_damping = linear_damping
    if angular_damping is not None:
        d.angular_damping = angular_damping
    if allow_sleep is not None:
        d.allow_sleep = allow_sleep
    if awake is not None:
        d.awake = awake
    if fixed_rotation is not None:
        d.fixed_rotation = fixed_rotation
    if bullet is not None:
        d.bullet = bullet
    if user_data is not None:
        d.user_data = user_data
    if int_user_data is not None:
        d.int_user_data = int_user_data
    return d


class _BodyDef(BodyDef):


    @property
    def user_data(self):
        if self._has_object_user_data():
            return self._get_object_user_data()
        else:
            return None

    @user_data.setter
    def user_data(self, ud):
        self._set_object_user_data(ud)



_classExtender(_BodyDef,['user_data'])








add_user_data_api(Body)
add_user_data_api(BodyDef)


class _Body( Body):






    @property
    def world(self):
        return self._get_world()

    def create_polygon_fixture(self, box=None,**kwargs):

        fixtureDef = b2FixtureDef()

        assert box is not None
        shape = b2PolygonShape()
        shape.setAsBox(box[0],box[1])

        fixtureDef.shape = shape 

        for kw in kwargs:
            setattr(fixtureDef,kw, kwargs[kw])

        return self.create_fixture(fixtureDef)

    def create_circle_fixture(self, radius, density=1, friction=0.2):
        fixtureDef = b2FixtureDef()

        shape = b2CircleShape()
        shape.radius = radius
        fixtureDef.friction = friction
        fixtureDef.density = density
        fixtureDef.shape = shape 
        return self.create_fixture(fixtureDef)

    def create_chain_fixture(self, vertices, prev_vertex, next_vertex, density=1, friction=0.2):
        fixtureDef = b2FixtureDef()
        shape = chain_shape(vertices=vertices,prev_vertex=prev_vertex, next_vertex=next_vertex)
        fixtureDef.shape = shape 
        return self.create_fixture(fixtureDef)



    def create_fixtures_from_shapes(self, shapes, density=1.0):
        if isinstance(shapes, b2Shape):
            shapes = [shapes]
        fixtures = []
        for shape in shapes:
            fixtureDef = b2FixtureDef()
            fixtureDef.density = density
            fixtureDef.shape = shape 
            fixtures.append( self.create_fixture(fixtureDef))
        return fixtures

    @property
    def type(self):
        return self.btype

    @property
    def next(self):
        if self._has_next():
            return self._get_next()
        else:
            return None

    @property
    def fixtures(self):
        flist = None
        if self._has_fixture_list():
            flist = self._get_fixture_list()
        return GenericB2dIter(flist)

    @property
    def joints(self):
        jlist = None
        if self._hasJointList():
            jlist = self._getJointList()
        return GenericB2dIter(jlist)

    @property
    def joint_list(self):
        return list(self.joints)
        
    @property
    def joint_list(self):
        return list(self.joints)





_classExtender(_Body,[ 'type','world','create_polygon_fixture','create_circle_fixture',
                      'create_chain_fixture',
                    'create_fixtures_from_shapes',
                     'next','fixtures','joints'], baseCls=Body)





import b2d
import numpy
from kivy.graphics import *


# class CanvasDraw(object):    
#     def __init__(self):
#         pass

#     def draw_solid_circle(self, center, radius, axis, color):
#         center = (numpy.array(center)-radius)
#         size = numpy.array([radius*2,radius*2])
#         #print "color",color
#         #with self.canvas:
#         e = Ellipse(pos=center,size=size,color=Color(*color))

#     def draw_segment(self,v1, v2, color):
#         #v1 = (numpy.array(v1)+self.offset)*self.scale
#         #v2 = (numpy.array(v2)+self.offset)*self.scale
#         #with self.canvas:
#         Line(points=[v1[0],v1[1],v2[0],v2[1]], width=1.0, color=Color(*color))

#     def draw_solid_polygon(self,vertices, color):
#         vertices = numpy.array(vertices)
#         vertices[:,0] #+= self.offset[0]
#         vertices[:,1] #+= self.offset[1]
#         #vertices*=self.scale

#         assert vertices.shape[0] ==vertexCount

#         v = [] 
#         indices = []
#         for i in range(vertices.shape[0]):
#             v.extend([vertices[i,0],vertices[i,1],0,0])
#             indices.append(i)
#         #with self.canvas:
#         Mesh(vertices=v,indices=indices,mode='triangle_fan',color=Color(*color))




class KivyBatchDebugDraw(b2d.batch_debug_draw_cls(True, True, False)):

    def __init__(self, flags=None):
        super(KivyBatchDebugDraw,self).__init__()

        # what is drawn
        if flags is None:
            flags = ['shape','joint', 'particle']#,'aabb','pair','center_of_mass','particle']
        self.flags = flags
        self.clear_flags(['shape','joint','aabb','pair','center_of_mass','particle'])
        for flag in flags:
            self.append_flags(flag)


    def draw_solid_polygons(self, points, sizes, colors):
        n_polygons = sizes.shape[0]
        start = 0
        for i in range(n_polygons):
            s = sizes[i]
            p = points[start:start+s,:]

            v = [] 
            indices = []
            for j in range(p.shape[0]):
                v.extend([p[j,0],p[j,1],0,0])
                indices.append(j)
            Mesh(vertices=v,indices=indices,mode='triangle_fan',color=Color(*colors[i,:]))
            start += s

    def draw_polygons(self, points, sizes, colors):
        n_polygons = sizes.shape[0]
        start = 0
        for i in range(n_polygons):
            s = sizes[i]
            p = points[start:start+s,:]

            Line(points = points, close=True, width = 1)

            start += s




    def draw_solid_circles(self, centers, radii, axis, colors):
        n = centers.shape[0]
        for i in range(n):
            radius = radii[i]
            center = centers[i,:]-radius
            size = ([radius*2,radius*2])
            e = Ellipse(pos=center,size=size,color=Color(*colors[i,:]))
        
    def draw_circles(self, centers, radii, colors):
        n = centers.shape[0]
        for i in range(n):
            radius = radii[i]
            Line(circle = (centers[i,0],centers[i,1], radii[i]), width = 1, color=Color(*colors[i,:]))
        

    def draw_points(self, centers, sizes, colors):
        pass

    def draw_segments(self, points, colors):
        n  = points.shape[0]
        for i in range(n):
            pass
            p =  points[i,0,0],points[i,0,1],points[i,1,0],points[i,1,1],
            Line(points=p, width=1.0, color=Color(*colors[i]))

    def draw_particles(self, centers, radius, colors=None):
        default_color = (1,1,1,1)

        n_particles = centers.shape[0]
        centers -= radius
        d = 2 * radius
        size = (d, d)
        PushMatrix()
        Translate(-radius, -radius, 0)
        for i in range(n_particles):

            if colors is None:
                c = default_color
            else:
                c = colors[i,:]
            Rectangle(size=size, pos=centers[i,:],  color=Color(*c))
        PopMatrix()



from .opencv_debug_draw import *
import matplotlib.pyplot as plt
import imageio


class GifGui(object):
    def __init__(self, testbed_cls, settings, testbed_kwargs=None):
        
        self.settings = settings
        self.testbed_cls = testbed_cls
        self.testbed_kwargs = testbed_kwargs
        self._testworld = None

        self._fps = settings.get('fps', 40.0)
        self._dt = 1.0 / self._fps
        self._t = settings.get('t',10)
        self._n = int(0.5 + self._t / self._dt)

        # settings
        resolution = settings.get("resolution",  (1000,1000))
        self.resolution = resolution

        self.image = numpy.zeros(list(self.resolution) + [3], dtype='uint8')
        self._image_list = []
        self.debug_draw = OpenCvBatchDebugDraw(image=self.image)
        self._filename = settings.get('filename', f"{str(testbed_cls.__name__)}.gif")
        self.debug_draw.screen_size = self.resolution
        self.debug_draw.flip_y = True
        self.debug_draw.scale =  settings.get("scale", 20.0)
        self.debug_draw.translate =settings.get("translate", (10,10))





    # run the world for a limited amount of steps
    def start_ui(self):
        
        self._testworld = self.testbed_cls(**self.testbed_kwargs)
        self._testworld.set_debug_draw(self.debug_draw)
        self._image_list = []

        for i in range(self._n):
            self._testworld.step(self._dt)
            self._testworld.draw_debug_data()

            self._image_list.append(self.image.copy())
            self.image[...] = 0

        imageio.mimsave(self._filename, self._image_list, fps=self._fps)



import b2d 

class EmptyDebugDraw(b2d.batch_debug_draw_cls(False, True, True)):

    def __init__(self):
        super(EmptyDebugDraw,self).__init__()


    def draw_solid_polygons(self, points, sizes, colors):
        pass

    def draw_polygons(self, points, sizes, colors):
        pass

    def draw_solid_circles(self, centers, radii, axis, colors):
        pass

    def draw_circles(self, centers, radii, colors):
        pass

    def _draw_circles(self, centers, radii,  colors, lw):
        pass

    def draw_points(self, centers, sizes, colors):
        pass

    def draw_segments(self, points, colors):
        pass

    def draw_particles(self, centers, radius, colors=None):
        pass


class NoGui(object):
    def __init__(self, testbed_cls, settings, testbed_kwargs=None):
        
        self.testbed_cls = testbed_cls
        self.testbed_kwargs = testbed_kwargs
        self._testworld = None

        self._dt = settings.get('dt', 1.0/40.0)
        self._t = settings.get('t',10)
        self._n = int(0.5 + self._t / self._dt) 
        self.debug_draw = EmptyDebugDraw()
        

    # run the world for a limited amount of steps
    def start_ui(self):
        
        self._testworld = self.testbed_cls(**self.testbed_kwargs)
        self._testworld.set_debug_draw(self.debug_draw)

        for i in range(self._n):
            self._testworld.step(self._dt)




from .kivy_debug_draw import *

# kivy-md
# from kivymd.app import MDApp
# from kivymd.uix.label import MDLabel
# from kivymd.app import MDApp

# kivy
from kivy.uix.widget import Widget
from kivy.graphics.instructions import *
from kivy.graphics.transformation import Matrix
from kivy.app import App
from kivy.uix.label import Label
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.scatter import Scatter
from kivy.clock import Clock
from kivy.config import Config
# from kivy.core.window import Window


class KivyWidget(Scatter):
    def __init__(self, kivy_gui, **kwargs):
        self.kivy_gui = kivy_gui
        super(KivyWidget, self).__init__(do_rotation=False)

        # dirty hack to resize window
        # Window.size = self.kivy_gui.resolution

        # clock to trigger stepping of world and rendering
        Clock.schedule_interval(self.step, self.kivy_gui._dt)


        # apply initial scale
        scale = self.kivy_gui.settings.get("scale", 30)
        self.apply_transform(Matrix().scale(scale, scale, scale),
                                     anchor=(0,0))

    def step(self, dt):
        self.canvas.clear()
        self.kivy_gui._testworld.step(dt)

        with self.canvas:
            self.kivy_gui._testworld.draw_debug_data()

    def handle_scroll(self, touch):
        factor = None
        if touch.button == 'scrolldown':
            if self.scale < self.scale_max:
                factor = 1.1
        elif touch.button == 'scrollup':
            if self.scale > self.scale_min:
                factor = 1 / 1.1
        if factor is not None:
            self.apply_transform(Matrix().scale(factor, factor, factor),
                                 anchor=touch.pos)
    def on_touch_down(self, touch):
        # Override Scatter's `on_touch_down` behavior for mouse scroll
        if touch.is_mouse_scrolling:
            self.handle_scroll(touch)
        else:
            world_pos = self.to_local(*touch.pos)
            handled_event = self.kivy_gui._testworld.on_mouse_down(world_pos)
            if not handled_event:
                super(KivyWidget, self).on_touch_down(touch)

    def on_touch_up(self, touch):
        if touch.is_mouse_scrolling:
            pass
        else:
            world_pos = self.to_local(*touch.pos)
            handled_event = self.kivy_gui._testworld.on_mouse_up(world_pos)
            if not handled_event:
                super(KivyWidget, self).on_touch_up(touch)

    def on_touch_move(self, touch):
        world_pos = self.to_local(*touch.pos)
        handled_event = self.kivy_gui._testworld.on_mouse_move(world_pos)
        if not handled_event:
            super(KivyWidget, self).on_touch_move(touch)

# this class implements the "GUI" interface for b2d
class KivyGui(App):
    def __init__(self, testbed_cls, settings, testbed_kwargs=None):
        


        super(KivyGui, self).__init__()

        self.settings = settings
        self.testbed_cls = testbed_cls
        self.testbed_kwargs = testbed_kwargs
        self._testworld = None

        self._fps = settings.get('fps', 40.0)
        self._dt = 1.0 / self._fps
        self._t = settings.get('t',10)
        self._n = int(0.5 + self._t / self._dt)

        # settings
        resolution = settings.get("resolution",  (1000,1000))
        self.resolution = resolution

        Config.set('graphics', 'width', f'{resolution[0]}')
        Config.set('graphics', 'height', f'{resolution[1]}')



        self.image = numpy.zeros(list(self.resolution) + [3], dtype='uint8')
        self._image_list = []
        self.debug_draw = KivyBatchDebugDraw()

        self.debug_draw.screen_size = self.resolution


    # 
    def start_ui(self):
        self._testworld = self.testbed_cls(**self.testbed_kwargs)
        self._testworld.set_debug_draw(self.debug_draw)

        # start the kivy ui
        self.run()



    def build(self):
        return KivyWidget(kivy_gui=self)




from . _b2d import *
from . _b2d import _World
from . tools import _classExtender, GenericB2dIter

from . extend_body import BodyDef
from . extend_shapes import polygon_shape
from . extend_fixture import fixture_def
from . extend_math import vec2
from . extend_collision import aabb
from . extend_joints import *
from . query_callback import QueryCallback



  


class World(_World):
    def __init__(self, gravity):
        super(World, self).__init__(gravity)
        self._batch_debug_draw_collector = None

    

    def set_destruction_listener(self, listener):
        self._set_destruction_listener(listener)


    def draw_debug_data(self, callback=None):
        self._draw_debug_data(callback)

    # helper functions
    def find_fixture(self, pos, margin=0.001):
        pos = vec2(pos)
        class AABBCallback(QueryCallback):
            def __init__(self, test_point):
                super(AABBCallback,self).__init__()

                self.test_point = vec2(test_point)
                self.fixture  = None

            def report_fixture(self, fixture):
                if fixture.test_point(self.test_point):
                    self.fixture = fixture
                    return False
                else:
                    return True
        box =  aabb(lower_bound=pos - vec2(margin, margin),
                        upper_bound=pos + vec2(margin, margin))

        query = AABBCallback(pos)
        self.query_aabb(query, box)
        return query.fixture 

    def find_body(self, pos, margin=0.001):
        fixture = self.find_fixture(pos=pos, margin=margin)
        if fixture is not None:
            return fixture.body
        else:
            return None

    def query_aabb_callback(self, **kwargs):
        

        class AABBCallback(QueryCallback):
            def __init__(self, callback):
                super(AABBCallback,self).__init__()
                self.callback = callback
            def report_fixture(self, fixture):
                return self.callback(fixture)

        callback = kwargs.pop('f')
        box = aabb(**kwargs)
        
        query = AABBCallback(callback)
        self.query_aabb(query, box)


    def find_closest_n_bodies(self, pos, r,r_fixture=None, n=None, body_filter=None):
        if r_fixture is None:
            r_fixture = r * 1.5  

        bodies_and_dists = []
        def f(fixture):
            body = fixture.body
            if body_filter is None or body_filter(body):
                center = body.world_center
                d = (pos-body.world_center).length
                if d < r: 
                    bodies_and_dists.append((body, d))
            return True

        self.query_aabb_callback(p=pos, r=r_fixture, f=f)
        sorted_bodies_and_dists = sorted(bodies_and_dists, key=lambda v :v[1])
        if n is not None and len(sorted_bodies_and_dists) >=n:
            return sorted_bodies_and_dists[0:n]
        else:
            return sorted_bodies_and_dists

    @property
    def bodies(self):
        blist = None
        if self.body_count > 0:
            blist = self._get_body_list()
        return GenericB2dIter(blist)

    @property
    def joints(self):
        blist = None
        if self.joint_count > 0:
            blist = self._get_joint_list()
        return GenericB2dIter(blist)

    @property
    def body_list(self):
        blist = None
        if self.body_count > 0:
            blist = self._get_body_list()
        return GenericB2dIter(blist)

    @property
    def joint_list(self):
        blist = None
        if self.joint_count > 0:
            blist = self._get_joint_list()
        return GenericB2dIter(blist)

    # functions
    def _create_body(self, type=None, body_def=None,shape=None, shapes=None, fixtures=None, shape_fixture=None, density = None,
                    **kwargs):

        if body_def is None:
            body_def = BodyDef()

        if type  is not None:
            body_def.type = type

        for attr_name,val in kwargs.items():
            setattr(body_def, attr_name, val)
      
        body = self._create_body_cpp(body_def)

    
        _shapes = []
        if shape is not None:
            _shapes.append(shape)
        if shapes is not None:
            if isinstance(shapes, Shape):
                shapes = [shapes]
            _shapes = _shapes + shapes


        for shape in _shapes:
            if shape_fixture is None:
                shape_fixture =fixture_def()
                if density is not None:
                    shape_fixture.density = density
            shape_fixture.shape = shape
            body.create_fixture(shape_fixture)


        if fixtures is not None:
            if isinstance(fixtures,FixtureDef):
                fixtures = [fixtures]
            for fixture in fixtures:
                body.create_fixture(fixture)

        return body

    def create_static_body(self, **kwargs):
        return self._create_body(type= BodyType.static, **kwargs)
    def create_dynamic_body(self, **kwargs):
        return self._create_body(type= BodyType.dynamic, **kwargs)
    def create_kinematic_body(self, **kwargs):
        return self._create_body(type= BodyType.kinematic, **kwargs)
    def create_body(self, *args, **kwargs):

        if(len(kwargs) == 0 and len(args) == 1 and isinstance(args[0], BodyDef)):
            return self._create_body_cpp(args[0])
        return self._create_body(*args, **kwargs)


    def create_box_body(self, box, density=None, **kwargs):
        fixture = fixture_def(density=density, shape=polygon_shape(box=box))
        return self._create_body(fixtures=fixture, **kwargs)



    def create_mouse_joint(self, *args,**kwargs):
        d = mouse_joint_def(*args, **kwargs)
        return self.create_joint(d)

    def create_prismatic_joint(self,*args,**kwargs):
        d = prismatic_joint_def(*args,**kwargs)
        return self.create_joint(d)

    def create_distance_joint(self,*args,**kwargs):
        d = distance_joint_def(*args,**kwargs)
        return self.create_joint(d)

    def create_revolute_joint(self,*args,**kwargs):
        d = revolute_joint_def(*args,**kwargs)
        return self.create_joint(d)
        
    def create_wheel_joint(self,*args,**kwargs):
        d = wheel_joint_def(*args,**kwargs)
        return self.create_joint(d)








def world(gravity = (0,-9.81)):
    return World(vec2(gravity))




from ..gif_gui import GifGui

import matplotlib.pyplot as plt
import matplotlib.animation as animation
import matplotlib
matplotlib.rc('animation', html='html5')

class MatplotlibGifGui(GifGui):
    def __init__(self, testbed_cls, settings, testbed_kwargs=None):
        t = settings.get('t', 10)
        settings['t'] = t

        fps = settings.get('fps', 24)
        settings['fps'] = t


        res = settings.get('resolution', (400,400))
        settings['resolution'] = res

        scale = settings.get('scale', 10)
        settings['scale'] = scale

        super(MatplotlibGifGui, self).__init__(testbed_cls=testbed_cls, settings=settings, testbed_kwargs=testbed_kwargs)






    # run the world for a limited amount of steps
    def start_ui(self):
        
        self._testworld = self.testbed_cls(**self.testbed_kwargs)
        self._testworld.set_debug_draw(self.debug_draw)
        
        self._image_list = []

        def make_next_img():
            self._testworld.step(self._dt)
            self._testworld.draw_debug_data()
            ret =  self.image.copy()
            self.image[...] = 0
            return ret 


        images = [make_next_img() for _ in range(self._n)]

        fig, ax = plt.subplots()
        imshow =  plt.imshow(images[0])

        def _update_plt(num):
            imshow.set_data(images[num])
            return imshow,


        dt = self._dt
        self.ani = animation.FuncAnimation(fig, _update_plt, 
                                          frames=self._n, 
                                          interval=dt * 1000.0,
                                          blit=True, 
                                          save_count=self._n,
                                          repeat_delay=500,
                                          cache_frame_data=False)

        return self.ani

        # return self._testworld, self


from ._b2d import *
from .extend_math import vec2
from .tools import _classExtender, GenericB2dIter
from . extend_user_data import add_user_data_api
from enum import Enum


add_user_data_api(JointDef)
add_user_data_api(Joint)



def _factory(cls, **kwargs):
    obj = cls()
    for k,v in kwargs.items():
        setattr(obj, k, v)
    return obj



def distance_joint_def(body_a, body_b, **kwargs):
    return _factory(DistanceJointDef,
        body_a=body_a, body_b=body_b, **kwargs)
    
def mouse_joint_def(body_a, body_b, **kwargs):
    return _factory(MouseJointDef,
        body_a=body_a, body_b=body_b, **kwargs)

def wheel_joint_def(body_a, body_b, **kwargs):
    return _factory(WheelJointDef,
        body_a=body_a, body_b=body_b, **kwargs)

def prismatic_joint_def(body_a, body_b, **kwargs):
    return _factory(PrismaticJointDef,
        body_a=body_a, body_b=body_b, **kwargs)

def revolute_joint_def(body_a, body_b, **kwargs):
    return _factory(RevoluteJointDef,
        body_a=body_a, body_b=body_b, **kwargs)



class _Joint(Joint):
    @property
    def next(self):
        if self._has_next():
            return self._get_next()
        else:
            return None


_classExtender(_Joint,['next'])





import inspect
import math

class GenericB2dIter(object):
    def __init__(self, currentItem):
        self.currentItem = currentItem
    def __next__(self):
        return self.next()
    def __iter__(self):
        return self
    def next(self):
        if self.currentItem is None:
            raise StopIteration
        else:
            c = self.currentItem
            self.currentItem = c.next
            return c

def _classExtender(moreCls, methods,baseCls=None):
    if baseCls is None:
        baseCls = inspect.getmro(moreCls)[1]
    for m in methods:
        setattr(baseCls, m, moreCls.__dict__[m])



from . _b2d import b2AABB
from . extend_math import vec2


Aabb = b2AABB


def aabb(lower_bound=None, upper_bound=None, p=None, r=None):
    if  lower_bound is not None and upper_bound is not None:

        lb = vec2(lower_bound)
        ub = vec2(upper_bound)
    elif r is not None and p is not None:
        p = vec2(p)
        rr = vec2(r,r)
        lb = p - rr
        ub = p + rr

    r = b2AABB()
    r.lower_bound = lb
    r.upper_bound = ub

    return r




def add_user_data_api(cls):

    @property
    def user_data(self):
        if self.has_object_user_data:
            return self._get_object_user_data()
        else:
            return None

    cls.user_data = user_data

    @user_data.setter
    def user_data(self, ud):
        if ud is None:
            self._clear_object_user_data()
        else:
            self._set_object_user_data(ud)
    cls.user_data = user_data



from . _b2d import *
from . extend_math import vec2
from . tools import _classExtender


def shape_filter(category_bits=None, mask_bits=None, group_index=None):
    f = Filter()
    if category_bits is not None:
        f.category_bits = category_bits
    if mask_bits is not None:
        f.mask_bits = mask_bits
    if group_index is not None:
        f.group_index = group_index


def two_sided_edge_shape(v0, v1):
    if BuildConfiguration.LIQUID_FUN:
        s = ChainShape()
        s.create_loop([
            v0,v1,v0
        ])
        return s
    else:
        shape = b2.EdgeShape()
        shape.set_two_sided(v0, v1)
        return shape;

# shape factories
def edge_shape(vertices):
    assert len(vertices) == 2
    s = EdgeShape()
    s.set_two_sided(vec2(vertices[0]),vec2(vertices[1]))
    return s

def chain_shape(vertices, prev_vertex=None, next_vertex=None):
    s = ChainShape()
    if  prev_vertex is None:
        prev_vertex = vertices[0]
    if  next_vertex is None:
        next_vertex = vertices[-1]

    prev_vertex = (float(prev_vertex[0]),float(prev_vertex[1]))
    next_vertex = (float(next_vertex[0]),float(next_vertex[1]))
    s.create_chain(vertices, prev_vertex, next_vertex)
    return s


def loop_shape(vertices):
    s = ChainShape()
    v = [vec2(vert) for vert in vertices]
    s.create_loop(v)


def circle_shape(radius, pos = (0,0)):
    s = CircleShape()
    s.radius = radius
    s.pos = vec2(pos)
    return s

def polygon_shape(box=None,center=(0,0),angle=0.0,vertices=None):
    s = PolygonShape()
    if vertices is None:
        s.set_as_box(box[0],box[1],center_x=center[0],center_y=center[1], angle=angle)
    else:
        verts = [ vec2(v) for v in vertices]
        s.set(verts)
    return s



def extend_polygon_shape():
    

    def _vertices(self):
        return [self.get_vertex(i) for i in range(self.vertex_count)]


    PolygonShape._vertices = _vertices
    PolygonShape.vertices = property(lambda self: self._vertices())


extend_polygon_shape()
del extend_polygon_shape





class _PolygonShape(PolygonShape):

    @property
    def vertices(self):
        return [self.getVertex(i) for i in range(self.vertexCount)]

    @vertices.setter
    def vertices(self, v):
        verts = [vec2(vert) for vert in  v]
        self.set(verts)
        

_classExtender(_PolygonShape, ['vertices'])







import numpy
import b2d 
import pygame
import pygame.locals 


class PyGameBatchDebugDraw(b2d.batch_debug_draw_cls(False, False, True)):

    def __init__(self, surface, flags=None):
        super(PyGameBatchDebugDraw,self).__init__()

        # what is drawn
        if flags is None:
            flags = ['shape','joint']#,'aabb','pair','center_of_mass','particle']
        self.flags = flags
        self.clear_flags(['shape','joint','aabb','pair','center_of_mass','particle'])
        for flag in flags:
            self.append_flags(flag)

        # the surface to draw on
        self._surface = surface

    def draw_solid_polygons(self, points, sizes, colors):
        self._draw_polygons(points, sizes, colors, 0)

    def draw_polygons(self, points, sizes, colors):
        self._draw_polygons(points, sizes, colors, 1)

    def _draw_polygons(self, points, sizes, colors, lw):

        n_polygons = sizes.shape[0]
        start = 0
        for i in range(n_polygons):
            s = sizes[i]
            p = points[start:start+s,:]
            pygame.draw.polygon(self._surface, colors[i,:], p, lw)
            start += s


    def draw_solid_circles(self, centers, radii, axis, colors):

        # draw circles itself
        self._draw_circles(centers, radii, colors, lw=0)

        p = centers - numpy.flip(axis,1) * radii[:, None] 

        n  = centers.shape[0]
        for i in range(n):
            pygame.draw.line(
                self._surface,
                (colors[i,:] * 0.5),
                centers[i,...],
                p[i,...]
            )

    def draw_circles(self, centers, radii, colors):
        self._draw_circles(centers, radii, colors, lw=1)

    def _draw_circles(self, centers, radii,  colors, lw):
        n = centers.shape[0]
        for i in range(n):
            pygame.draw.circle(self._surface, 
                colors[i,:], 
                centers[i,:],
                radii[i], 
                lw)

    def draw_points(self, centers, sizes, colors):
        pass

    def draw_segments(self, points, colors):
        n  = points.shape[0]
        for i in range(n):
            pygame.draw.line(
                self._surface,
                colors[i,:],
                points[i,0,:],
                points[i,1,:]
            )

    def draw_particles(self, centers, radius, colors=None):
        default_color = (255,255,255,255)

        n_particles = centers.shape[0]
        centers -= radius
        d = 2 * radius
        for i in range(n_particles):

            if colors is None:
                c = default_color
            else:
                c = colors[i,:]

            pygame.draw.rect(
                self._surface, c,
                (centers[i,0], centers[i,1], d, d)
            )




from . _b2d import *


class QueryCallback(QueryCallbackCaller):

    def __init__(self):
        super(QueryCallback,self).__init__(self)

    def report_fixture(self, fixture):
        raise NotImplementedError 
    def report_particle(self, particleSystem, index):
        return False
    def should_query_particle_system(self, particleSystem):
        return False




from ._b2d import *
import numpy 







def vec2(*args):
    l = len(args)
    if l == 1:
        a = args[0]
        if isinstance(a,Vec2):
            return a
        else:
            return Vec2(float(a[0]),float(a[1]))
    elif l == 2:
        return Vec2(float(args[0]),float(args[1]))

def transform(position = (0,0), rotation = 0.0):
    rot = b2Rot(float(rotation))
    return b2Transform(vec2(position), rot)

class Vec2Iter(object):
    def __init__(self,vector, currentIndex=0):
        self.vector = vector
        self.currentIndex = currentIndex
    def __next__(self):
        return self.next()
    def __iter__(self):
        return self
    def next(self):
        if self.currentIndex==2:
            raise StopIteration
        else:
            c = self.vector[self.currentIndex]
            self.currentIndex +=1
            return c



def extendB2Vec2():
    

    def __setitem__(self, key, val):

        if int(key) == 0:
            self.x = val
        elif int(key) == 1:
            self.y = val
        else:
            raise RuntimeError("wrong index %s"%str(key))
    Vec2.__setitem__ = __setitem__


    def __getitem__(self, key):

        if int(key) == 0:
            return self.x
        elif int(key) == 1:
            return self.y
        else:
            raise RuntimeError("wrong index %s"%str(key))
    Vec2.__getitem__ = __getitem__

    def isfinite(self):
        return numpy.isfinite(self.x) and numpy.isfinite(self.y)
    Vec2.isfinite = isfinite
  

    def __iter__(self):
        return Vec2Iter(self)
    Vec2.__iter__ = __iter__

  
    def as_tuple(self):
        return (self.x,self.y)
    Vec2.as_tuple = as_tuple

    def __repr__(self):
        return "(%f,%f)"%(self.x,self.y)
    Vec2.__repr__ = __repr__

    def __str__(self):
        return "(%f,%f)"%(self.x,self.y)
    Vec2.__str__ = __str__
extendB2Vec2()
del extendB2Vec2



import sys
import glob
import os
from pathlib import Path

from pybind11 import get_cmake_dir,get_include
# Available at setup time due to pyproject.toml
from pybind11.setup_helpers import Pybind11Extension, build_ext
from setuptools import setup, find_packages

__version__ = "0.3.3"

# The main interface is through Pybind11Extension.
# * You can add cxx_std=11/14/17, and then build_ext can be removed.
# * You can set include_pybind11=false to add the include directory yourself,
#   say from a submodule.
#
# Note:
#   Sort input source files if you glob sources to ensure bit-for-bit
#   reproducible builds (https://github.com/pybind/python_example/pull/53)

liquidfun = True

binding_sources = [
    "python/src/main.cpp",
    "python/src/def_build_config.cpp",
    "python/src/b2Body.cxx",
    "python/src/b2Collision.cxx",
    "python/src/b2Contact.cxx",
    "python/src/b2Draw.cxx",
    "python/src/b2Fixture.cxx",
    "python/src/b2Joint.cxx",
    "python/src/b2JointDef.cxx",
    "python/src/b2Math.cxx",
    "python/src/b2Shape.cxx",
    "python/src/b2WorldCallbacks.cxx",
    "python/src/b2World.cxx"
]

include_dirs = [
    "external/pybind11-2.8.1/include/"
]

if liquidfun:
    binding_sources += [
        "python/src/b2Particle.cxx",
        "python/src/b2ParticleGroup.cxx",
        "python/src/b2ParticleSystem.cxx",
        "python/src/pyEmitter.cxx",
        "python/src/extensions/b2Emitter.cpp"
    ]
    macros = [
        ('PYBOX2D_LIQUID_FUN',1),
    ]
    base_dir = "external/box2d-ecf398ca73f31b282cf9e6a500d8af6665654617"
    

else:
    base_dir = "external/box2d-2.4.1"
    macros = []

box2d_include_dirs = [
    os.path.join(base_dir,"include"),
    os.path.join(base_dir,"src"),
]
src_dir = os.path.join(base_dir,"src")
box2d_sources = sorted([str(path) for path in Path(src_dir).rglob('*.cpp')])



ext_modules = [
    Pybind11Extension("b2d._b2d",

        binding_sources + box2d_sources,
        include_dirs=box2d_include_dirs+include_dirs,
        # Example: passing in the version to the compiled code
        define_macros = [('VERSION_INFO', __version__)] + macros,
        ),
]


install_requires = [
    "numpy"
]
setup(
    name="b2d",
    version=__version__,
    author="Thorsten Beier",
    author_email="derthorstenbeier@gmail.com",
    url="https://github.com/pybind/python_example",
    description="A test project using pybind11",
    long_description="",
    ext_modules=ext_modules,
    packages=find_packages(where='./python/module', exclude="test"),
    install_requires=install_requires,
    extras_require={"test": "pytest"},
    package_dir = {'': 'python/module'},
    # Currently, build_ext only provides an optional "highest supported C++
    # level" feature, but in the future it may provide more features.
    cmdclass={"build_ext": build_ext},
    zip_safe=False,
    python_requires=">=3.6",
)


from ._b2d import *
from .tools import _classExtender





def fixture_def(shape=None, friction=None, restitution=None, density=None, isSensor=None, shape_filter=None, user_data=None, group_index=None):
    fd = FixtureDef()
    if shape is not None : 
        fd.shape = shape
    if friction is not None : 
        fd.friction = friction
    if restitution is not None : 
        fd.restitution = restitution
    if density is not None : 
        fd.density = density
    if isSensor is not None : 
        fd.isSensor = isSensor

    if group_index is not None:
        if shape_filter is None:
            shape_filter = b2Filter()
        shape_filter.group_index = int(group_index)

    if shape_filter is not None : 
        fd.filter = shape_filter

    if user_data is not None:
        fd.user_data = user_data
    return fd


class _FixtureDef(FixtureDef):

    @property
    def user_data(self):
        if self._has_user_data():
            return self._get_user_data()
        else:
            return None
    @user_data.setter
    def user_data(self, ud):
        if self._has_user_data():
            return self._delete_user_data()
        self._set_user_data(ud)


    @property
    def shape(self):
        return self._shape
    @shape.setter
    def shape(self, shape):
        self._set_shape(shape)

_classExtender(_FixtureDef,['user_data','shape'])


    
class _Fixture(Fixture):

    @property
    def user_data(self):
        if self._has_user_data():
            return self._get_user_data()
        else:
            return None
    @user_data.setter
    def user_data(self, ud):
        if self._has_user_data():
            return self._delete_user_data()
        self._set_user_data(ud)

    @property
    def next(self):
        if self._hasNext():
            return self._getNext()
        else:
            return None
    @property
    def shape(self):
        return self._get_shape()
_classExtender(_Fixture,['user_data','next','shape'])




import numpy
import b2d 
import cv2 as cv

class OpenCvBatchDebugDraw(b2d.batch_debug_draw_cls(False, True, True)):

    def __init__(self, image, flags=None):
        super(OpenCvBatchDebugDraw,self).__init__()

        # what is drawn
        if flags is None:
            flags = ['shape','joint','particle']#,'aabb','pair','center_of_mass','particle']
        self.flags = flags
        self.clear_flags(['shape','joint','aabb','pair','center_of_mass','particle'])
        for flag in flags:
            self.append_flags(flag)

        # the image to draw on
        self._image = image


    def draw_solid_polygons(self, points, sizes, colors):
        self._draw_polygons(points, sizes, colors, True)

    def draw_polygons(self, points, sizes, colors):
        self._draw_polygons(points, sizes, colors, False)

    def _draw_polygons(self, points, sizes, colors, fill):
        line_type = 8
        n_polygons = sizes.shape[0]
        start = 0
        for i in range(n_polygons):
            s = sizes[i]
            p = points[start:start+s,:].astype('int32')
            color = tuple(map(int, colors[i,:]))
            if fill:
                cv.fillPoly(self._image, [p], color, line_type)
            else:
                cv.polylines(self._image, [p], True, color, line_type)
            start += s



    def draw_solid_circles(self, centers, radii, axis, colors):
        self._draw_circles(centers, radii, colors, lw=-1)
        
    def draw_circles(self, centers, radii, colors):
        self._draw_circles(centers, radii, colors, lw=1)

    def _draw_circles(self, centers, radii,  colors, lw):
        line_type = 8
        thickness = 1
        n = centers.shape[0]
        for i in range(n):
            color = tuple(map(int, colors[i,:]))
            cv.circle(self._image,
                       centers[i,:].astype('int32'),
                       radii[i].astype('int32'),
                       color,
                       lw,
                       line_type)


    def draw_points(self, centers, sizes, colors):
        pass

    def draw_segments(self, points, colors):
        line_type = 8
        thickness = 1

        n  = points.shape[0]
        for i in range(n):
            color = tuple(map(int, colors[i,:]))
            cv.line(self._image,
                points[i,0,:].astype('int32'),
                points[i,1,:].astype('int32'),  
                color)


    def draw_particles(self, centers, radius, colors=None):
    
        default_color = (255,255,255,255)

        n_particles = centers.shape[0]
        centers -= radius
        d = 2 * radius
        for i in range(n_particles):

            if colors is None:
                color = default_color
            else:
                color = tuple(map(int, colors[i,:]))

            p0 = (centers[i,:] - radius).astype('int32')
            p1 = (p0 + 2*radius).astype('int32')
            cv.rectangle(self._image, p0, p1, color,-1)



import b2d
import numpy
import random
import time





class DebugDrawCallback(object):

    def __init__(self, pre_debug_draw, post_debug_draw):
        self._pre_debug_draw = pre_debug_draw
        self._post_debug_draw = post_debug_draw

    def pre_debug_draw(self):
        self._pre_debug_draw()

    def post_debug_draw(self ):
        self._post_debug_draw()




# class TestbedSettings(object):
#     def __init__(self, 
#         resolution=(1000,1000), 
#         fps=40, 
#         ppm=30,
#         offset=(0,0)
#         =)

class TestbedBase(
    b2d.DestructionListener,
    b2d.ContactListener
    
):
    @classmethod
    def run(cls, gui_factory, gui_settings=None, testbed_kwargs=None):
        if gui_settings is None:
            gui_settings = dict()
        if testbed_kwargs is None:
            testbed_kwargs = dict()        

        ui = gui_factory(testbed_cls=cls, testbed_kwargs=testbed_kwargs, settings=gui_settings)

        return ui.start_ui()

    def __init__(self, gravity=b2d.vec2(0,-9.81)):
        b2d.ContactListener.__init__(self)
        b2d.DestructionListener.__init__(self)

        # Box2D-related
        self.points = []
        self.world = None
        self.bomb = None
        self.mouse_joint = None

        # self.framework_settings = FrameworkSettings
        self.step_count = 0
        self.is_paused = False
        self.__time_last_step = None
        self.current_fps = 0.0

        # the b2d world itself
        self.world = b2d.world(gravity)
        self.groundbody = self.world.create_body()
        
        self.world.set_contact_listener(self)
        self.world.set_destruction_listener(self)
        self.iter = 0

        # debug draw 
        self.debug_draw = None

    def set_debug_draw(self, debug_draw):
        if self.debug_draw is not None:
            raise RuntimeError("debug draw has already been set") 
        self.debug_draw = debug_draw
        self.world.set_debug_draw(self.debug_draw)

    def pre_debug_draw(self):
        pass
    def post_debug_draw(self):
        pass

    def draw_debug_data(self):

        callback = DebugDrawCallback(
            pre_debug_draw=self.pre_debug_draw,
            post_debug_draw=self.post_debug_draw
        )
        self.world.draw_debug_data(callback)


    def set_gui(self, gui):
        self._gui = gui

    def is_key_down(self, key):
        return self._gui.is_key_down(key)


    def step(self, dt):
        self.pre_step(dt)
        self.world.step(dt, 3, 3)
        if self.__time_last_step is None:
            self.__time_last_step  = time.time()
        else:
            t_now = time.time()
            dt = t_now - self.__time_last_step 
            self.__time_last_step = t_now
            try:
                self.current_fps = 1.0/dt
            except ZeroDivisionError:
                self.current_fps = float('inf')
                
        self.step_count += 1
        self.iter += 1
        self.post_step(dt)


    def say_goodbye_world(self):
        pass


    def pre_step(self, dt):
        pass

    def post_step(self, dt):
        pass
            

    def get_particle_parameter_value(self):
        return 0



    def on_mouse_move(self, p):
        """
        Mouse moved to point p, in world coordinates.
        """
        if self.mouse_joint is not None:
            self.mouse_joint.target = p
            return True
        else:
            return False

    def on_mouse_down(self, p):
        """
        Indicates that there was a left click at point p (world coordinates)
        """

       
        if self.mouse_joint is not None:
            self.world.destroy_joint(self.mouse_joint)
            self.mouse_joint = None
            return False


        body = self.world.find_body(pos=p)
        if body is not None:
            
            kwargs = dict(
                body_a=self.groundbody,
                body_b=body,
                target=p,
                max_force=50000.0 * body.mass,
                stiffness=1000.0
            )

            self.mouse_joint = self.world.create_mouse_joint(**kwargs)
            body.awake = True

        return body is not None

    def on_mouse_up(self, p):
        """
        Left mouse button up.
        """

        if self.mouse_joint is not None:
            self.world.destroy_joint(self.mouse_joint)
            self.mouse_joint = None
            return True
        else:
            return False

    def on_key_down(self, key):
        return False

    def on_key_up(self, key):
        return False

    # # ContactListener
    # def begin_contact(self, contact):
    #     pass

    # def end_contact(self, contact):
    #     pass

    # def begin_contact_particle_body(self, particleSystem, particleBodyContact):
    #     pass

    # def begin_contact_particle(self, particleSystem, indexA, indexB):
    #     pass

    # def end_contact_particle(self, particleSystem, indexA, indexB):
    #     pass

    # def pre_solve(self, contact, oldManifold):
    #     pass

    # def post_solve(self, contact, impulse):
    #    pass

    # DestructionListener
    def say_goodbye_joint(self, joint):
        pass

    def say_goodbye_fixture(self, fixture):
        pass
    def say_goodbye_particle_group(self, particleGroup):
        pass
    def say_goodbye_particle_system(self, particleSystem,index):
        pass





import b2d



class JupyterBatchDebugDraw(b2d.batch_debug_draw_cls(False, False, True)):

    def __init__(self, canvas, flags=None):
        super(JupyterBatchDebugDraw,self).__init__()

        # what is drawn
        if flags is None:
            flags = ['shape','joint','aabb','pair','center_of_mass','particle']
        self.flags = flags
        self.clear_flags(['shape','joint','aabb','pair','center_of_mass','particle'])
        for flag in flags:
            self.append_flags(flag)

        # the canvas to draw on
        self._canvas = canvas

    def draw_solid_polygons(self, points, sizes, colors):
        self._canvas.fill_styled_polygons(
            points=points.ravel(),
            points_per_polygon=sizes,
            color=colors.ravel(),
            alpha=1.0
        )

    def draw_polygons(self, points, sizes, colors):
        self._canvas.stroke_styled_polygons(
            points=points.ravel(),
            points_per_polygon=sizes,
            color=colors.ravel(),
            alpha=1.0
        )

    def draw_solid_circles(self, centers, radii, axis, colors):
        
        # ignore axis atm
        self._canvas.fill_styled_circles(
            centers[:,0],
            centers[:,1],
            radii,
            colors.ravel(),
            1.0
        )
        
    def draw_circles(self, centers, radii, colors):
        self._canvas.stroke_styled_circles(
            centers[:,0],
            centers[:,1],
            radii,
            color=colors.ravel(),
            alpha=1.0
        )

    def draw_points(self, centers, sizes, colors):
        self._canvas.stroke_styled_circles(
            centers,
            sizes,
            color=colors.ravel(),
            alpha=1.0
        )

    def draw_segments(self, points, colors):
        self._canvas.stroke_styled_line_segments(
            points=points,
            color=colors.ravel(),
            alpha=1.0
        )


    def draw_particles(self, centers, radius, colors=None):



        if colors is None:

            # print("draw_particles",centers.shape, radius)
            r = radius
            self._canvas.save()
            self._canvas.translate(x=-r, y=-r)
            self._canvas.fill_style = "rgba(0,100,255,1)" 
            self._canvas.fill_rects(centers[:,0], centers[:,1], r*2)
            self._canvas.restore()

            # for r_mult,a in zip([2], [0.5]):
            #     r = radius * r_mult
            #     self._canvas.save()
            #     self._canvas.translate(x=-r, y=-r)
            #     self._canvas.fill_style = f"rgba(0,255,0,{a*0.7})" 
            #     self._canvas.fill_rects(centers[:,0], centers[:,1], r*2)
            #     self._canvas.restore()

        else:
            alpha = (colors[:,3])/255.0
            colors = colors[:,0:3]

            r = radius
            self._canvas.save()
            self._canvas.translate(x=-r, y=-r)
            self._canvas.fill_styled_rects(centers[:,0], centers[:,1], r*2, r*2, colors, alpha)
            self._canvas.restore()


     
        # self._canvas.fill_style = "rgba(0,255,0,0.3)" 
        # self._canvas.fill_rects(centers[:,0], centers[:,1], radius*2 * 1.2)

        # centers += radius 
        # centers -= radius * 8
        # self._canvas.fill_style = "rgba(0,255,0,0.01)" 
        # self._canvas.fill_rects(centers[:,0], centers[:,1], radius*16)

